@@ -13,12 +13,9 @@
 #include <sys/stat.h>
 #include <time.h>
 
-#ifdef _MSC_VER
-#pragma warning(disable: 4786)
-#endif
-
 #include <string>
 #include <vector>
+#include <set>
 #include <map>
 
 #if defined(__unix__)
@@ -52,6 +49,7 @@
 #endif
 
 #include "Scintilla.h"
+#include "ILexer.h"
 
 #include "GUI.h"
 
@@ -65,6 +63,8 @@
 #include "SciTE.h"
 #include "Mutex.h"
 #include "JobQueue.h"
+#include "Cookie.h"
+#include "Worker.h"
 #include "SciTEBase.h"
 
 
@@ -105,7 +105,7 @@
 
 // extract the next RTF control word from *style
 void GetRTFNextControl(char **style, char *control) {
-	int len;
+	ptrdiff_t len;
 	char *pos = *style;
 	*control = '\0';
 	if ('\0' == *pos) return;
@@ -240,10 +240,10 @@ void SciTEBase::SaveToRTF(FilePath saveName, int start, int end) {
 				} else {
 					strcat(lastStyle, RTF_SETBACKGROUND "1");	// Default back
 				}
-				if (sd.specified & StyleDefinition::sdBold) {
-					strcat(lastStyle, sd.bold ? RTF_BOLD_ON : RTF_BOLD_OFF);
+				if (sd.specified & StyleDefinition::sdWeight) {
+					strcat(lastStyle, sd.IsBold() ? RTF_BOLD_ON : RTF_BOLD_OFF);
 				} else {
-					strcat(lastStyle, defaultStyle.bold ? RTF_BOLD_ON : RTF_BOLD_OFF);
+					strcat(lastStyle, defaultStyle.IsBold() ? RTF_BOLD_ON : RTF_BOLD_OFF);
 				}
 				if (sd.specified & StyleDefinition::sdItalics) {
 					strcat(lastStyle, sd.italics ? RTF_ITALIC_ON : RTF_ITALIC_OFF);
@@ -436,7 +436,7 @@ void SciTEBase::SaveToHTML(FilePath saveName) {
 					sd.font = sdmono.font;
 					sd.size = sdmono.size;
 					sd.italics = sdmono.italics;
-					sd.bold = sdmono.bold;
+					sd.weight = sdmono.weight;
 				}
 
 				if (sd.specified != StyleDefinition::sdNone) {
@@ -448,7 +448,7 @@ void SciTEBase::SaveToHTML(FilePath saveName) {
 					if (sd.italics) {
 						fprintf(fp, "\tfont-style: italic;\n");
 					}
-					if (sd.bold) {
+					if (sd.IsBold()) {
 						fprintf(fp, "\tfont-weight: bold;\n");
 					}
 					if (wysiwyg && sd.font.length()) {
@@ -733,20 +733,20 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 	class PDFObjectTracker {
 	private:
 		FILE *fp;
-		int *offsetList, tableSize;
+		long *offsetList, tableSize;
 	public:
 		int index;
 		PDFObjectTracker(FILE *fp_) {
 			fp = fp_;
 			tableSize = 100;
-			offsetList = new int[tableSize];
+			offsetList = new long[tableSize];
 			index = 1;
 		}
 		~PDFObjectTracker() {
 			delete []offsetList;
 		}
 		void write(const char *objectData) {
-			unsigned int length = strlen(objectData);
+			size_t length = strlen(objectData);
 			// note binary write used, open with "wb"
 			fwrite(objectData, sizeof(char), length, fp);
 		}
@@ -759,8 +759,8 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 		int add(const char *objectData) {
 			// resize xref offset table if too small
 			if (index > tableSize) {
-				int newSize = tableSize * 2;
-				int *newList = new int[newSize];
+				long newSize = tableSize * 2;
+				long *newList = new long[newSize];
 				for (int i = 0; i < tableSize; i++) {
 					newList[i] = offsetList[i];
 				}
@@ -777,17 +777,17 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 			return index++;
 		}
 		// builds xref table, returns file offset of xref table
-		int xref() {
+		long xref() {
 			char val[32];
 			// xref start index and number of entries
-			int xrefStart = ftell(fp);
+			long xrefStart = ftell(fp);
 			write("xref\n0 ");
 			write(index);
 			// a xref entry *must* be 20 bytes long (PDF1.4Ref(p64))
 			// so extra space added; also the first entry is special
 			write("\n0000000000 65535 f \n");
 			for (int i = 0; i < index - 1; i++) {
-				sprintf(val, "%010d 00000 n \n", offsetList[i]);
+				sprintf(val, "%010ld 00000 n \n", offsetList[i]);
 				write(val);
 			}
 			return xrefStart;
@@ -816,7 +816,7 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 		PDFStyle *style;
 		int fontSize;		// properties supplied by user
 		int fontSet;
-		int pageWidth, pageHeight;
+		long pageWidth, pageHeight;
 		GUI::Rectangle pageMargin;
 		//
 		PDFRender() {
@@ -901,7 +901,7 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 			int pagesRef = pageObjectStart + pageCount;
 			for (int i = 0; i < pageCount; i++) {
 				sprintf(buffer, "<</Type/Page/Parent %d 0 R\n"
-				        "/MediaBox[ 0 0 %d %d"
+				        "/MediaBox[ 0 0 %ld %ld"
 				        "]\n/Contents %d 0 R\n"
 				        "/Resources %d 0 R\n>>\n",
 				        pagesRef, pageWidth, pageHeight,
@@ -921,10 +921,10 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 			sprintf(buffer, "<</Type/Catalog/Pages %d 0 R >>\n", pagesRef);
 			int catalogRef = oT->add(buffer);
 			// append the cross reference table (PDF1.4Ref(p64))
-			int xref = oT->xref();
+			long xref = oT->xref();
 			// end the file with the trailer (PDF1.4Ref(p67))
 			sprintf(buffer, "trailer\n<< /Size %d /Root %d 0 R\n>>"
-			        "\nstartxref\n%d\n%%%%EOF\n",
+			        "\nstartxref\n%ld\n%%%%EOF\n",
 			        oT->index, catalogRef, xref);
 			oT->write(buffer);
 		}
@@ -1065,19 +1065,19 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 	propItem = props.GetExpanded("export.pdf.margins");
 	ps = StringDup(propItem.c_str());
 	next = GetNextPropItem(ps, buffer, 32);
-	if (0 >= (pr.pageMargin.left = atol(buffer))) {
+	if (0 >= (pr.pageMargin.left = static_cast<int>(atol(buffer)))) {
 		pr.pageMargin.left = PDF_MARGIN_DEFAULT;
 	}
 	next = GetNextPropItem(next, buffer, 32);
-	if (0 >= (pr.pageMargin.right = atol(buffer))) {
+	if (0 >= (pr.pageMargin.right = static_cast<int>(atol(buffer)))) {
 		pr.pageMargin.right = PDF_MARGIN_DEFAULT;
 	}
 	next = GetNextPropItem(next, buffer, 32);
-	if (0 >= (pr.pageMargin.top = atol(buffer))) {
+	if (0 >= (pr.pageMargin.top = static_cast<int>(atol(buffer)))) {
 		pr.pageMargin.top = PDF_MARGIN_DEFAULT;
 	}
 	GetNextPropItem(next, buffer, 32);
-	if (0 >= (pr.pageMargin.bottom = atol(buffer))) {
+	if (0 >= (pr.pageMargin.bottom = static_cast<int>(atol(buffer)))) {
 		pr.pageMargin.bottom = PDF_MARGIN_DEFAULT;
 	}
 	delete []ps;
@@ -1099,7 +1099,7 @@ void SciTEBase::SaveToPDF(FilePath saveName) {
 
 		if (sd.specified != StyleDefinition::sdNone) {
 			if (sd.italics) { pr.style[i].font |= 2; }
-			if (sd.bold) { pr.style[i].font |= 1; }
+			if (sd.IsBold()) { pr.style[i].font |= 1; }
 			if (sd.fore.length()) {
 				getPDFRGB(pr.style[i].fore, sd.fore.c_str());
 			} else if (i == STYLE_DEFAULT) {
@@ -1210,7 +1210,7 @@ static void defineTexStyle(StyleDefinition &style, FILE* fp, int istyle) {
 		fputs("\\textit{", fp);
 		closing_brackets++;
 	}
-	if (style.bold) {
+	if (style.IsBold()) {
 		fputs("\\textbf{", fp);
 		closing_brackets++;
 	}
