@@ -5,6 +5,7 @@
 // Copyright 1998-2007 by Neil Hodgson <neilh@scintilla.org>
 // The License.txt file describes the conditions under which this software may be distributed.
 
+#include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -13,13 +14,12 @@
 #include <time.h>
 #include <locale.h>
 
-#ifdef _MSC_VER
-#pragma warning(disable: 4786)
-#endif
-
 #include <string>
+#include <vector>
+#include <set>
 #include <map>
 #include <sstream>
+#include <algorithm>
 
 #if defined(GTK)
 
@@ -49,17 +49,55 @@ inline bool IsASpace(unsigned int ch) {
     return (ch == ' ') || ((ch >= 0x09) && (ch <= 0x0d));
 }
 
+static std::set<std::string> FilterFromString(std::string values) {
+	std::set<std::string> fs;
+	std::istringstream isValues(values);
+	while (!isValues.eof()) {
+		std::string sValue;
+		isValues >> sValue;
+		if (!sValue.empty())
+			fs.insert(sValue);
+	}
+	return fs;
+}
+
+void ImportFilter::SetFilter(std::string sExcludes, std::string sIncludes) {
+	excludes = FilterFromString(sExcludes);
+	includes = FilterFromString(sIncludes);
+}
+
+bool ImportFilter::IsValid(std::string name) const {
+	if (!includes.empty()) {
+		return includes.count(name) > 0;
+	} else {
+		return excludes.count(name) == 0;
+	}
+}
+
 bool PropSetFile::caseSensitiveFilenames = false;
 
 PropSetFile::PropSetFile(bool lowerKeys_) : lowerKeys(lowerKeys_), superPS(0) {
 }
 
+PropSetFile::PropSetFile(const PropSetFile &copy) : lowerKeys(copy.lowerKeys), props(copy.props), superPS(copy.superPS) {
+}
+
 PropSetFile::~PropSetFile() {
 	superPS = 0;
 	Clear();
 }
 
-void PropSetFile::Set(const char *key, const char *val, int lenKey, int lenVal) {
+PropSetFile &PropSetFile::operator=(const PropSetFile &assign) {
+	if (this != &assign) {
+		lowerKeys = assign.lowerKeys;
+		superPS = assign.superPS;
+		props = assign.props;
+		enumnext = "";
+	}
+	return *this;
+}
+
+void PropSetFile::Set(const char *key, const char *val, ptrdiff_t lenKey, ptrdiff_t lenVal) {
 	if (!*key)	// Empty keys are not supported
 		return;
 	if (lenKey == -1)
@@ -118,17 +156,86 @@ bool PropSetFile::Exists(const char *key) const {
 }
 
 SString PropSetFile::Get(const char *key) const {
-	mapss::const_iterator keyPos = props.find(std::string(key));
-	if (keyPos != props.end()) {
-		return SString(keyPos->second.c_str());
-	} else {
-		if (superPS) {
-			// Failed here, so try in base property set
-			return superPS->Get(key);
-		} else {
-			return "";
+	const std::string sKey(key);
+	const PropSetFile *psf = this;
+	while (psf) {
+		mapss::const_iterator keyPos = psf->props.find(sKey);
+		if (keyPos != psf->props.end()) {
+			return SString(keyPos->second.c_str());
+		}
+		// Failed here, so try in base property set
+		psf = psf->superPS;
+	}
+	return "";
+}
+
+static SString ShellEscape(const char *toEscape) {
+	SString str(toEscape);
+	for (int i = static_cast<int>(str.length()-1); i >= 0; --i) {
+		switch (str[i]) {
+		case ' ':
+		case '|':
+		case '&':
+		case ',':
+		case '`':
+		case '"':
+		case ';':
+		case ':':
+		case '!':
+		case '^':
+		case '$':
+		case '{':
+		case '}':
+		case '(':
+		case ')':
+		case '[':
+		case ']':
+		case '=':
+		case '<':
+		case '>':
+		case '\\':
+		case '\'':
+			str.insert(i, "\\");
+			break;
+		default:
+			break;
 		}
 	}
+	return str.c_str();
+}
+
+SString PropSetFile::Evaluate(const char *key) const {
+	if (strchr(key, ' ')) {
+		if (isprefix(key, "escape ")) {
+			SString val = Get(key+7);
+			return ShellEscape(val.c_str());
+		} else if (isprefix(key, "star ")) {
+			const std::string sKeybase(key + 5);
+			// Create set of variables with values
+			mapss values;
+			// For this property set and all base sets
+			for (const PropSetFile *psf = this; psf; psf = psf->superPS) {
+				mapss::const_iterator it = psf->props.lower_bound(sKeybase);
+				while ((it != psf->props.end()) && (it->first.find(sKeybase) == 0)) {
+					mapss::iterator itDestination = values.find(it->first);
+					if (itDestination == values.end()) {
+						// Not present so add
+						values[it->first] = it->second;
+					}
+					++it;
+				}
+			}
+			// Concatenate all variables
+			std::string combination;
+			for (mapss::const_iterator itV = values.begin(); itV != values.end(); ++itV) {
+				combination += itV->second;
+			}
+			return SString(combination.c_str());
+		}
+	} else {
+		return Get(key);
+	}
+	return "";
 }
 
 // There is some inconsistency between GetExpanded("foo") and Expand("$(foo)").
@@ -165,7 +272,7 @@ static int ExpandAllInPlace(const PropSetFile &props, SString &withVars, int max
 		}
 
 		SString var(withVars.c_str(), varStart + 2, varEnd);
-		SString val = props.Get(var.c_str());
+		SString val = props.Evaluate(var.c_str());
 
 		if (blankVars.contains(var.c_str())) {
 			val.clear(); // treat blankVar as an empty string (e.g. to block self-reference)
@@ -236,7 +343,7 @@ char *PropSetFile::ToString() const {
 /**
  * Get a line of input. If end of line escaped with '\\' then continue reading.
  */
-static bool GetFullLine(const char *&fpc, int &lenData, char *s, int len) {
+static bool GetFullLine(const char *&fpc, size_t &lenData, char *s, size_t len) {
 	bool continuation = true;
 	s[0] = '\0';
 	while ((len > 1) && lenData > 0) {
@@ -278,8 +385,30 @@ static bool IsCommentLine(const char *line) {
 	return (*line == '#');
 }
 
+bool IsPropertiesFile(const FilePath &filename) {
+	FilePath ext = filename.Extension();
+	if (EqualCaseInsensitive(ext.AsUTF8().c_str(), PROPERTIES_EXTENSION + 1))
+		return true;
+	return false;
+}
+
+static bool GenericPropertiesFile(const FilePath &filename) {
+	std::string name = filename.BaseName().AsUTF8();
+	if (name == "abbrev" || name == "Embedded")
+		return true;
+	return name.find("SciTE") != std::string::npos;
+}
+
+void PropSetFile::Import(FilePath filename, FilePath directoryForImports, const ImportFilter &filter, std::vector<FilePath> *imports) {
+	if (Read(filename, directoryForImports, filter, imports)) {
+		if (imports && (std::find(imports->begin(),imports->end(), filename) == imports->end())) {
+			imports->push_back(filename);
+		}
+	}
+}
+
 bool PropSetFile::ReadLine(const char *lineBuffer, bool ifIsTrue, FilePath directoryForImports,
-                           FilePath imports[], int sizeImports) {
+                           const ImportFilter &filter, std::vector<FilePath> *imports) {
 	//UnSlash(lineBuffer);
 	if (!IsSpaceOrTab(lineBuffer[0]))    // If clause ends with first non-indented line
 		ifIsTrue = true;
@@ -288,17 +417,24 @@ bool PropSetFile::ReadLine(const char *lineBuffer, bool ifIsTrue, FilePath direc
 		ifIsTrue = GetInt(expr) != 0;
 	} else if (isPrefix(lineBuffer, "import ") && directoryForImports.IsSet()) {
 		SString importName(lineBuffer + strlen("import") + 1);
-		importName += ".properties";
-		FilePath importPath(directoryForImports, FilePath(GUI::StringFromUTF8(importName.c_str())));
-		if (Read(importPath, directoryForImports, imports, sizeImports)) {
-			if (imports) {
-				for (int i = 0; i < sizeImports; i++) {
-					if (!imports[i].IsSet()) {
-						imports[i] = importPath;
-						break;
-					}
+		if (importName == "*") {
+			// Import all .properties files in this directory except for system properties
+			FilePathSet directories;
+			FilePathSet files;
+			directoryForImports.List(directories, files);
+			for (size_t i = 0; i < files.size(); i ++) {
+				FilePath fpFile = files[i];
+				if (IsPropertiesFile(fpFile) &&
+					!GenericPropertiesFile(fpFile) &&
+					filter.IsValid(fpFile.BaseName().AsUTF8())) {
+					FilePath importPath(directoryForImports, fpFile);
+					Import(importPath, directoryForImports, filter, imports);
 				}
 			}
+		} else if (filter.IsValid(importName.c_str())) {
+			importName += ".properties";
+			FilePath importPath(directoryForImports, FilePath(GUI::StringFromUTF8(importName.c_str())));
+			Import(importPath, directoryForImports, filter, imports);
 		}
 	} else if (ifIsTrue && !IsCommentLine(lineBuffer)) {
 		Set(lineBuffer);
@@ -306,8 +442,8 @@ bool PropSetFile::ReadLine(const char *lineBuffer, bool ifIsTrue, FilePath direc
 	return ifIsTrue;
 }
 
-void PropSetFile::ReadFromMemory(const char *data, int len, FilePath directoryForImports,
-                                 FilePath imports[], int sizeImports) {
+void PropSetFile::ReadFromMemory(const char *data, size_t len, FilePath directoryForImports,
+                                 const ImportFilter &filter, std::vector<FilePath> *imports) {
 	const char *pd = data;
 	char lineBuffer[60000];
 	bool ifIsTrue = true;
@@ -320,12 +456,12 @@ void PropSetFile::ReadFromMemory(const char *data, int len, FilePath directoryFo
 				}
 			}
 		}
-		ifIsTrue = ReadLine(lineBuffer, ifIsTrue, directoryForImports, imports, sizeImports);
+		ifIsTrue = ReadLine(lineBuffer, ifIsTrue, directoryForImports, filter, imports);
 	}
 }
 
 bool PropSetFile::Read(FilePath filename, FilePath directoryForImports,
-                       FilePath imports[], int sizeImports) {
+                       const ImportFilter &filter, std::vector<FilePath> *imports) {
 	FILE *rcfile = filename.Open(fileRead);
 	if (rcfile) {
 		char propsData[60000];
@@ -336,7 +472,7 @@ bool PropSetFile::Read(FilePath filename, FilePath directoryForImports,
 			data += 3;
 			lenFile -= 3;
 		}
-		ReadFromMemory(data, lenFile, directoryForImports, imports, sizeImports);
+		ReadFromMemory(data, lenFile, directoryForImports, filter, imports);
 		return true;
 	}
 	return false;
@@ -388,49 +524,51 @@ static bool startswith(const std::string &s, const char *keybase) {
 }
 
 SString PropSetFile::GetWildUsingStart(const PropSetFile &psStart, const char *keybase, const char *filename) {
-	mapss::iterator it = props.lower_bound(std::string(keybase));
-	while ((it != props.end()) && startswith(it->first, keybase)) {
-		const char *orgkeyfile = it->first.c_str() + strlen(keybase);
-		char *keyptr = NULL;
+	const std::string sKeybase(keybase);
+	const size_t lenKeybase = strlen(keybase);
+	const PropSetFile *psf = this;
+	while (psf) {
+		mapss::const_iterator it = psf->props.lower_bound(sKeybase);
+		while ((it != psf->props.end()) && startswith(it->first, keybase)) {
+			const char *orgkeyfile = it->first.c_str() + lenKeybase;
+			char *keyptr = NULL;
 
-		if (strncmp(orgkeyfile, "$(", 2) == 0) {
-			const char *cpendvar = strchr(orgkeyfile, ')');
-			if (cpendvar) {
-				SString var(orgkeyfile, 2, cpendvar-orgkeyfile);
-				SString s = psStart.GetExpanded(var.c_str());
-				keyptr = StringDup(s.c_str());
+			if (strncmp(orgkeyfile, "$(", 2) == 0) {
+				const char *cpendvar = strchr(orgkeyfile, ')');
+				if (cpendvar) {
+					SString var(orgkeyfile, 2, cpendvar-orgkeyfile);
+					SString s = psStart.GetExpanded(var.c_str());
+					keyptr = StringDup(s.c_str());
+				}
 			}
-		}
-		const char *keyfile = keyptr;
+			const char *keyfile = keyptr;
 
-		if (keyfile == NULL)
-			keyfile = orgkeyfile;
+			if (keyfile == NULL)
+				keyfile = orgkeyfile;
 
-		for (;;) {
-			const char *del = strchr(keyfile, ';');
-			if (del == NULL)
-				del = keyfile + strlen(keyfile);
-			if (MatchWild(keyfile, del - keyfile, filename, caseSensitiveFilenames)) {
-				delete []keyptr;
+			for (;;) {
+				const char *del = strchr(keyfile, ';');
+				if (del == NULL)
+					del = keyfile + strlen(keyfile);
+				if (MatchWild(keyfile, del - keyfile, filename, caseSensitiveFilenames)) {
+					delete []keyptr;
+					return SString(it->second.c_str());
+				}
+				if (*del == '\0')
+					break;
+				keyfile = del + 1;
+			}
+			delete []keyptr;
+
+			if (0 == strcmp(it->first.c_str(), keybase)) {
 				return SString(it->second.c_str());
 			}
-			if (*del == '\0')
-				break;
-			keyfile = del + 1;
+			++it;
 		}
-		delete []keyptr;
-
-		if (0 == strcmp(it->first.c_str(), keybase)) {
-			return SString(it->second.c_str());
-		}
-		++it;
-	}
-	if (superPS) {
-		// Failed here, so try in super property set
-		return static_cast<PropSetFile *>(superPS)->GetWildUsingStart(psStart, keybase, filename);
-	} else {
-		return "";
+		// Failed here, so try in base property set
+		psf = psf->superPS;
 	}
+	return "";
 }
 
 SString PropSetFile::GetWild(const char *keybase, const char *filename) {
@@ -446,7 +584,7 @@ SString PropSetFile::GetNewExpand(const char *keybase, const char *filename) {
 	while (cpvar && (maxExpands > 0)) {
 		const char *cpendvar = strchr(cpvar, ')');
 		if (cpendvar) {
-			int lenvar = cpendvar - cpvar - 2;  	// Subtract the $()
+			ptrdiff_t lenvar = cpendvar - cpvar - 2;  	// Subtract the $()
 			char *var = StringDup(cpvar + 2, lenvar);
 			SString val = GetWild(var, filename);
 			if (0 == strcmp(var, keybase))
@@ -739,7 +877,7 @@ int SString::search(const char *sFind, lenpos_t start) const {
 	if (start < sLen) {
 		const char *sFound = strstr(s + start, sFind);
 		if (sFound) {
-			return sFound - s;
+			return static_cast<int>(sFound - s);
 		}
 	}
 	return -1;
@@ -781,16 +919,16 @@ char *SContainer::StringAllocate(lenpos_t len) {
 	}
 }
 
-char *SContainer::StringAllocate(const char *s, lenpos_t len) {
-	if (s == 0) {
+char *SContainer::StringAllocate(const char *sValue, lenpos_t len) {
+	if (sValue == 0) {
 		return 0;
 	}
 	if (len == measure_length) {
-		len = strlen(s);
+		len = strlen(sValue);
 	}
 	char *sNew = new char[len + 1];
 	if (sNew) {
-		memcpy(sNew, s, len);
+		memcpy(sNew, sValue, len);
 		sNew[len] = '\0';
 	}
 	return sNew;
