@@ -199,7 +199,7 @@ bool StringList::InList(const char *s) {
  * The length of the word to compare is passed too.
  * Letter case can be ignored or preserved (default).
  */
-const char *StringList::GetNearestWord(const char *wordStart, int searchLen, bool ignoreCase /*= false*/, SString wordCharacters /*='/0' */, int wordIndex /*= -1 */) {
+const char *StringList::GetNearestWord(const char *wordStart, size_t searchLen, bool ignoreCase /*= false*/, SString wordCharacters /*='/0' */, int wordIndex /*= -1 */) {
 	int start = 0; // lower bound of the api array block to search
 	int end = len - 1; // upper bound of the api array block to search
 	int pivot; // index of api array element just being compared
@@ -314,7 +314,7 @@ static unsigned int LengthWord(const char *word, char otherSeparator) {
 			endWord--;
 		}
 	}
-	return endWord - word;
+	return static_cast<unsigned int>(endWord - word);
 }
 
 /**
@@ -329,7 +329,7 @@ static unsigned int LengthWord(const char *word, char otherSeparator) {
  */
 char *StringList::GetNearestWords(
     const char *wordStart,
-    int searchLen,
+    size_t searchLen,
     bool ignoreCase /*= false*/,
     char otherSeparator /*= '\0'*/,
     bool exactLen /*=false*/) {
