@@ -170,7 +170,8 @@ bool SciTEWin::ModelessHandler(MSG *pmsg) {
 
 //  DoDialog is a bit like something in PC Magazine May 28, 1991, page 357
 int SciTEWin::DoDialog(HINSTANCE hInst, const TCHAR *resName, HWND hWnd, DLGPROC lpProc) {
-	int result = ::DialogBoxParam(hInst, resName, hWnd, lpProc, reinterpret_cast<LPARAM>(this));
+	int result = static_cast<int>(
+		::DialogBoxParam(hInst, resName, hWnd, lpProc, reinterpret_cast<LPARAM>(this)));
 
 	if (result == -1) {
 		GUI::gui_string errorNum = GUI::StringFromInteger(::GetLastError());
@@ -309,7 +310,7 @@ bool SciTEWin::SaveAsDialog() {
 void SciTEWin::SaveACopy() {
 	FilePath path = ChooseSaveName(filePath.Directory(), "Save a Copy");
 	if (path.IsSet()) {
-		SaveBuffer(path);
+		SaveBuffer(path, true);
 	}
 }
 
@@ -539,7 +540,7 @@ void SciTEWin::Print(
 	                           ptDpi.y, 72);
 	HFONT fontHeader = ::CreateFontA(headerLineHeight,
 	                                0, 0, 0,
-	                                sdHeader.bold ? FW_BOLD : FW_NORMAL,
+	                                sdHeader.weight,
 	                                sdHeader.italics,
 	                                sdHeader.underlined,
 	                                0, 0, 0,
@@ -557,7 +558,7 @@ void SciTEWin::Print(
 	                           ptDpi.y, 72);
 	HFONT fontFooter = ::CreateFontA(footerLineHeight,
 	                                0, 0, 0,
-	                                sdFooter.bold ? FW_BOLD : FW_NORMAL,
+	                                sdFooter.weight,
 	                                sdFooter.italics,
 	                                sdFooter.underlined,
 	                                0, 0, 0,
@@ -815,6 +816,13 @@ static void FillComboFromProps(HWND combo, PropSetFile &props) {
 	}
 }
 
+void SciTEWin::ShowBackgroundProgress(const GUI::gui_string &explanation, int size, int progress) {
+	backgroundStrip.visible = !explanation.empty();
+	SizeSubWindows();
+	if (backgroundStrip.visible)
+		backgroundStrip.SetProgress(explanation, size, progress);
+}
+
 class DialogFindReplace : public Dialog, public SearchUI  {
 	bool advanced;
 public:
@@ -861,7 +869,7 @@ void DialogFindReplace::FillFields() {
 		SetCheck(pSearcher->reverseFind ? IDDIRECTIONUP : IDDIRECTIONDOWN, true);
 	}
 	if (advanced) {
-		SetCheck(IDFINDSTYLE, pSearcher->findInStyle);
+		SetCheck(IDFINDINSTYLE, pSearcher->findInStyle);
 	}
 }
 
@@ -890,23 +898,23 @@ BOOL SciTEWin::FindMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			wFindReplace.Destroy();
 			return FALSE;
-		} else if ( (ControlIDOfCommand(wParam) == IDOK) ||
-		            (ControlIDOfCommand(wParam) == IDMARKALL) ) {
+		} else if ( (ControlIDOfWParam(wParam) == IDOK) ||
+		            (ControlIDOfWParam(wParam) == IDMARKALL) ) {
 			dlg.GrabFields();
 			if (closeFind) {
 				::EndDialog(hDlg, IDOK);
 				wFindReplace.Destroy();
 			}
-			if (ControlIDOfCommand(wParam) == IDMARKALL){
+			if (ControlIDOfWParam(wParam) == IDMARKALL){
 				MarkAll();
 			}
-			FindNext(reverseFind);
+			FindNext(reverseFind ^ IsKeyDown(VK_SHIFT));
 			return TRUE;
-		} else if (ControlIDOfCommand(wParam) == IDFINDINSTYLE) {
+		} else if (ControlIDOfWParam(wParam) == IDFINDINSTYLE) {
 			if (FindReplaceAdvanced()) {
 				findInStyle = dlg.Checked(IDFINDINSTYLE);
 				dlg.Enable(IDFINDSTYLE, findInStyle);
@@ -922,7 +930,7 @@ BOOL CALLBACK SciTEWin::FindDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM l
 	return Caller(hDlg, message, lParam)->FindMessage(hDlg, message, wParam);
 }
 
-BOOL SciTEWin::HandleReplaceCommand(int cmd) {
+BOOL SciTEWin::HandleReplaceCommand(int cmd, bool reverseFind) {
 	if (!wFindReplace.GetID())
 		return TRUE;
 	HWND hwndFR = reinterpret_cast<HWND>(wFindReplace.GetID());
@@ -935,7 +943,7 @@ BOOL SciTEWin::HandleReplaceCommand(int cmd) {
 
 	int replacements = 0;
 	if (cmd == IDOK) {
-		FindNext(false);
+		FindNext(reverseFind);
 	} else if (cmd == IDREPLACE) {
 		ReplaceOnce();
 	} else if ((cmd == IDREPLACEALL) || (cmd == IDREPLACEINSEL)) {
@@ -976,20 +984,20 @@ BOOL SciTEWin::ReplaceMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			props.Set("Replacements", "");
 			UpdateStatusBar(false);
 			::EndDialog(hDlg, IDCANCEL);
 			wFindReplace.Destroy();
 			return FALSE;
-		} else if (ControlIDOfCommand(wParam) == IDFINDINSTYLE) {
+		} else if (ControlIDOfWParam(wParam) == IDFINDINSTYLE) {
 			if (FindReplaceAdvanced()) {
 				findInStyle = dlg.Checked(IDFINDINSTYLE);
 				dlg.Enable(IDFINDSTYLE, findInStyle);
 			}
 			return TRUE;
 		} else {
-			return HandleReplaceCommand(ControlIDOfCommand(wParam));
+			return HandleReplaceCommand(ControlIDOfWParam(wParam), IsKeyDown(VK_SHIFT));
 		}
 	}
 
@@ -1148,12 +1156,12 @@ BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			wFindInFiles.Destroy();
 			return FALSE;
 
-		} else if (ControlIDOfCommand(wParam) == IDOK) {
+		} else if (ControlIDOfWParam(wParam) == IDOK) {
 			findWhat = dlg.ItemTextU(IDFINDWHAT);
 			props.Set("find.what", findWhat.c_str());
 			memFinds.Insert(findWhat.c_str());
@@ -1179,12 +1187,12 @@ BOOL SciTEWin::GrepMessage(HWND hDlg, UINT message, WPARAM wParam) {
 			} else {
 				return FALSE;
 			}
-		} else if (ControlIDOfCommand(wParam) == IDDOTDOT) {
+		} else if (ControlIDOfWParam(wParam) == IDDOTDOT) {
 			FilePath directory(dlg.ItemTextG(IDDIRECTORY));
 			directory = directory.Directory();
 			dlg.SetItemText(IDDIRECTORY, directory.AsInternal());
 
-		} else if (ControlIDOfCommand(wParam) == IDBROWSE) {
+		} else if (ControlIDOfWParam(wParam) == IDBROWSE) {
 
 			// This code was copied and slightly modifed from:
 			// http://www.bcbdev.com/faqs/faq62.htm
@@ -1329,10 +1337,10 @@ BOOL SciTEWin::GoLineMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			return FALSE;
-		} else if (ControlIDOfCommand(wParam) == IDOK) {
+		} else if (ControlIDOfWParam(wParam) == IDOK) {
 			BOOL bHasLine;
 			int lineNumber = static_cast<int>(
 			                     ::GetDlgItemInt(hDlg, IDGOLINE, &bHasLine, FALSE));
@@ -1390,10 +1398,10 @@ BOOL SciTEWin::AbbrevMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			return FALSE;
-		} else if (ControlIDOfCommand(wParam) == IDOK) {
+		} else if (ControlIDOfWParam(wParam) == IDOK) {
 			SString sAbbrev = dlg.ItemTextU(IDABBREV);
 			strncpy(abbrevInsert, sAbbrev.c_str(), sizeof(abbrevInsert));
 			abbrevInsert[sizeof(abbrevInsert) - 1] = '\0';
@@ -1445,11 +1453,11 @@ BOOL SciTEWin::TabSizeMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			return FALSE;
-		} else if ((ControlIDOfCommand(wParam) == IDCONVERT) ||
-			(ControlIDOfCommand(wParam) == IDOK)) {
+		} else if ((ControlIDOfWParam(wParam) == IDCONVERT) ||
+			(ControlIDOfWParam(wParam) == IDOK)) {
 			BOOL bOK;
 			int tabSize = static_cast<int>(::GetDlgItemInt(hDlg, IDTABSIZE, &bOK, FALSE));
 			if (tabSize > 0)
@@ -1459,10 +1467,10 @@ BOOL SciTEWin::TabSizeMessage(HWND hDlg, UINT message, WPARAM wParam) {
 				wEditor.Call(SCI_SETINDENT, indentSize);
 			bool useTabs = static_cast<bool>(::IsDlgButtonChecked(hDlg, IDUSETABS));
 			wEditor.Call(SCI_SETUSETABS, useTabs);
-			if (ControlIDOfCommand(wParam) == IDCONVERT) {
+			if (ControlIDOfWParam(wParam) == IDCONVERT) {
 				ConvertIndentation(tabSize, useTabs);
 			}
-			::EndDialog(hDlg, ControlIDOfCommand(wParam));
+			::EndDialog(hDlg, ControlIDOfWParam(wParam));
 			return TRUE;
 		}
 	}
@@ -1521,13 +1529,13 @@ BOOL SciTEWin::ParametersMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			if (!modalParameters) {
 				wParameters.Destroy();
 			}
 			return FALSE;
-		} else if (ControlIDOfCommand(wParam) == IDOK) {
+		} else if (ControlIDOfWParam(wParam) == IDOK) {
 			ParamGrab();
 			::EndDialog(hDlg, IDOK);
 			if (!modalParameters) {
@@ -1621,10 +1629,10 @@ BOOL SciTEWin::AboutMessage(HWND hDlg, UINT message, WPARAM wParam) {
 		break;
 
 	case WM_COMMAND:
-		if (ControlIDOfCommand(wParam) == IDOK) {
+		if (ControlIDOfWParam(wParam) == IDOK) {
 			::EndDialog(hDlg, IDOK);
 			return TRUE;
-		} else if (ControlIDOfCommand(wParam) == IDCANCEL) {
+		} else if (ControlIDOfWParam(wParam) == IDCANCEL) {
 			::EndDialog(hDlg, IDCANCEL);
 			return FALSE;
 		}
